{
           Grand Theft Auto: Vice City
            Language File (GXT) Reader
                     v1.0

         Written by Xentor (Xentor@usa.net)

     Uses HashTables library written by Tommi Johtela
     (See header of HashTables.pas for copyright info)

Usage:
  Create an instance of GXTReader and load a GXT file.
  myReader := GXTReader.Create;
  myReader.loadGXT('C:\Games\GTAVC\TEXT\american.gxt');

  The MAIN sub-table is always accessible.  To read from another, call
  selectTable.  Only one sub-table (Besides MAIN) can be loaded at a time, just
  like in the game code itself.

  To retrieve a string, call getRef() with the key name.  Keys are 1-8
  characters and contain uppercase letters, numbers, and underscores(_) only.

Algorithms:
  The loading routine is entirely sequential.  The only significant drain comes
  from a quicksort of each sub-table before the strings are loaded.  Since all
  data is stored in hash tables, getRef() access time is constant, regardless of
  filesize.  The selectTable procedure utilizes a binary search among the 100+
  tables for maximum efficiency.  Both loading and access times were nearly
  instantaneous on my P3-1GHz WinXP machine.

Notes:
  'american.gxt' is used as an example above, but note that getRef returns a
  unicode string.  This class fully supports multiple languages with no
  performance loss.
}

unit GXTAccess;

interface

uses SysUtils, StrUtils, Classes, HashTables, Contnrs;

type

  TableEntry = record
    name : String[8];
    addr : longInt;
  end;

  GXTReader = class
  public
    //Load a GXT file from disk
    procedure loadGXT(fileName : String);

    //Select which sub-table is being used
    // All searches check MAIN first, then the active sub-table
    procedure selectTable(tableName : String);

    //Provide a key (8 char max), and receive the corresponding GXT entry, or '' if the key is not found.
    function getRef(key: String) : String;
  private
    procedure loadTable(fStream : TFileStream; tableidx : integer; tableName : String);
    function readUnicode(fStream : TFileStream; numbytes : integer) : WideString;
    function readChars(fStream : TFileStream; numChars : integer) : String;
    procedure sort( var r : Array of TableEntry; lo, up : integer );
  end;

var

  lstTables : Array of TStringTable;  //Array of sub-tables (Hash tables)
  lstTableNames : Array of String[8]; //Array of names of sub-tables, corresponds to lstTables
  iNumTables : integer; // Number of sub-tables
  mainTable : TStringTable; //MAIN table, stored seperately from lstTables for efficiency
  currentTable: TStringTable; //References the current sub-table in use (selectTable)

implementation

  //Returns a string of up to 8 ANSI characters from a file stream.  File pointer will advance
  function GXTReader.readChars(fStream : TFileStream; numChars : integer) : String;
  var
    buf : Array[1..8] of Char;
  begin
    //Read the chars into a char array, then move to a string, to avoid Delphi's idiocy.
    fStream.ReadBuffer(buf,numChars);
    Result := TrimRight(LeftStr(buf,numChars));
  end;

  //Read a unicode string from a file.  numbytes must be an even number.  File pointer will advance
  function GXTReader.readUnicode(fStream : TFileStream; numbytes : integer) : WideString;
  var
    buf : Array[1..50] of Char;
    sBuf : String;
    wBuf : WideString;
    charsLeft : integer;
    idx : integer;
    ch1,ch2 : integer;
  begin
    // Read blocks of 50 bytes (To keep Delphi happy), and append to sBuf
    sBuf := '';  charsLeft := numbytes;
    while charsLeft > 50 do
    begin
      fStream.ReadBuffer(buf,50);
      sBuf := sBuf + buf;
      charsLeft := charsLeft - 50;
    end;
    fStream.ReadBuffer(buf,charsLeft);
    sBuf := sBuf + buf;

    //Step through sBuf by twos, calculating each unicode char as a sum of two bytes manually
    //Remember, a unicode value is equal to byte1 + 256(byte2) in little-endian.
    idx := 1;
    while idx < numbytes do
    begin
      ch1 := Ord(sBuf[idx]);
      ch2 := Ord(sBuf[idx+1]);
      wBuf := wBuf + Chr(ch1 + (256 * ch2));
      idx := idx + 2;
    end;
    Result := TrimRight(wBuf);
  end;

  //Called by loadGXT to read in each sub-table seperately.
  //tableidx corresponds to the index in lstTables and lstTableNames
  //tableName is not used yet, but could be used to verify the correct file position
  procedure GXTReader.loadTable(fStream : TFileStream; tableidx : integer; tableName : String);
  var
    sName : String[8];  //Placeholder to read name of table
    sKeyword : String[4]; //Placeholder to read TKEY/TDAT (Can be used for error-checking
    iNumKeys : integer; //Number of key/value pairs in this sub-table
    iKeyblockSize : longInt; //Size of TKEY block, not including header
    iDatblockSize : longInt; //Size of TDAT block, not including header
    keys : array of TableEntry; //Array of key/address pairs
    table : TStringTable;
    iDatblockStart : integer; //Address of first byte of first TDAT value, to calc relative addresses

    offset : integer;
    idx : integer;
    sTmp : WideString;
    blocksize : integer;
  begin
    //MAIN table does not have a tablename header, so don't look for one.
    if tableName <> 'MAIN' then sName := readChars(fStream,8);

    //Read in TKEY header
    sKeyword := readChars(fStream,4); //TKEY
    fStream.ReadBuffer(iKeyblockSize,4);

    //Each key is a 4-byte address and an 8-byte string, so divide
    // block size by 12 (4+8) to get number of keys.
    iNumKeys := iKeyBlockSize div 12;
    setLength(keys,iNumKeys); //Resize key array

    //Read in address/key pairs
    for idx := 0 to iNumKeys - 1 do
    begin
      fStream.readBuffer(keys[idx].addr,4);
      keys[idx].name := readChars(fStream,8);
    end;

    //Use quicksort algorithm to sort key array by address.
    // (Initially sorted by key name)
    sort(keys,0,iNumKeys-1);

    //Read in TDAT header
    sKeyword := readChars(fStream,4); //TDAT
    fStream.readBuffer(iDatblockSize,4);
    iDatblockStart := fStream.Position;

    //Get a handle on the hash table being used, and set its capacity
    if tableName = 'MAIN' then
      table := mainTable
    else
      table := lstTables[tableidx];
    table.Capacity := iNumKeys;

    //Loop through the keys, grab their values, and hash them.
    for idx := 0 to iNumKeys - 1 do
    begin
      //First, make sure we're in the right place.  Some values have an extra
      // 00 00 at the end, so we might have to advance a byte or two.
      offset := keys[idx].addr - (fStream.Position - iDatblockStart);
      if offset > 0 then fStream.Seek(offset,soFromCurrent);

      //For most values, just calculate the size by subtracting this address from
      // the next one.  For the last address, use the TDAT size instead.
      if idx < (iNumKeys - 1) then
        blocksize := keys[idx+1].addr - keys[idx].addr
      else
        blocksize := iDatblockSize - keys[idx].addr;

      //Read in the unicode string, and add it to the hash table.
      sTmp := readUnicode(fStream,blocksize);
      table.insert(keys[idx].name,sTmp);
    end;

  end;

  procedure GXTReader.loadGXT(fileName : String);
  var
    fStream : TFileStream;
    sKeyword: String[4];
    iTableSize : longInt;
    tables : Array of TableEntry;

    idx : integer;
  begin
    //Open the file and grab the TABL header.
    fStream := TFileStream.Create(fileName,fmOpenRead);
    sKeyword := readChars(fStream,4);

    //Table headers are 4-byte addresses and 8-byte strings, so divide the
    // table size (Second word) by 12 to get the number of subtables.
    //Initialize all of the arrays to the proper size.
    fstream.ReadBuffer(iTableSize,4);
    iNumTables := iTableSize div 12;
    SetLength(tables, iNumTables);
    SetLength(lstTables,iNumTables);
    SetLength(lstTableNames,iNumTables);

    //Read in the entire TABL section... Easy, huh?
    for idx := 0 to iNumTables - 1 do
    begin
      tables[idx].name := readChars(fStream,8);
      fStream.ReadBuffer(tables[idx].addr,4);
    end;

    //For each sub-table, initialize a hash, set the lstTableNames entry,
    // advance the file pointer to the proper location, and load the table.
    for idx := 0 to iNumTables - 1 do
    begin
      lstTables[idx] := TStringTable.Create;
      lstTableNames[idx] := tables[idx].name;
      fStream.Seek(tables[idx].addr,soFromBeginning);
      loadTable(fStream,idx,tables[idx].name);
    end;
    //To avoid null references, start currentTable off as a duplicate of MAIN
    currentTable := mainTable;
    fStream.Destroy;
  end;

  procedure GXTReader.selectTable(tableName : String);
  var min,max,idx : integer;
  begin
    if iNumTables = 0 then exit;
    // Run a binary search on the list of tables, and
    //  set currentTable to lstTables[tableName] if possible
    min := 0;
    max := iNumTables;
    idx := (max + min) div 2;
    while (max > min) and (lstTableNames[idx] <> tableName) do
    begin
      if lstTableNames[idx] < tableName then
        min := idx + 1
      else
        max := idx - 1;
      idx := (max + min) div 2;
    end;

    if min <= max then currentTable := lstTables[idx];
  end;

  function GXTReader.getRef(key : String) : String;
  var sTmp : WideString;
  begin
    { If mainTable contains key, return value.
      otherwise, if currentTable contains key, return that value
      Otherwise, return "" }

    sTmp := mainTable.Items[key];
    if sTmp = '' then sTmp := currentTable.Items[key];
    Result := sTmp;
  end;


  //QuickSort implementation adapted from:
  // http://www.dcc.uchile.cl/~rbaeza/handbook/algs/4/413.sort.p.html
  // (Too lazy to code it from scratch)
  procedure GXTReader.sort( var r : Array of TableEntry; lo, up : integer );

     var  i, j : integer;
          tempr : TableEntry;
     begin
     while up>lo do begin
          i := lo;
          j := up;
          tempr := r[lo];
          {*** Split file in two ***}
          while i<j do begin
               while r[j].addr > tempr.addr do
                    j := j-1;
               r[i] := r[j];
               while (i<j) and (r[i].addr <= tempr.addr) do
                    i := i+1;
               r[j] := r[i]
               end;
          r[i] := tempr;
          {*** Sort recursively ***}
          sort(r,lo,i-1);
          lo := i+1
          end
     end;


initialization
  //Hey, this is as good a place to initialize mainTable as any...
  mainTable := TStringTable.Create;
  //I know... It looks so lonely down here... Hence the pointless comments!
end.
